# Paravirtualized guest support
#
# Initially derived from virt-inst, Copyright 2005-2006 Red Hat, Inc.
# Copyright 2006-2007 Novell, Inc.
# Author: Charles Coffing <ccoffing@novell.com>
#
# This software may be freely redistributed under the terms of the GNU
# general public license.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA 

import os
import re

from .VMDefaults import VMDefaults
from .xen_guest import *
from .exceptions import *
from .globals import *
from .log import log
from vmdisks import Disk
from uuid import UUID
from . import hypervisor
from . import mount
from . import paths
from . import util
import vmdisks
import time
import grub.GrubConf

def loopsetup(vdev):
    """Sets up the loop mapping for a disk image.
    """
    i = 0
    while True:
        ldev = '/dev/loop%i' % (i)
        if not os.path.exists(ldev):
            if i == 0:
                os.system("losetup -f > /dev/null 2>&1")
                if os.path.exists(ldev):
                    continue
            break
        i += 1
        fd = os.popen("losetup %s '%s' 2> /dev/null" % (ldev, vdev))
        if not fd.close():
            log.debug("losetup %s '%s'" % (ldev, vdev))
            return ldev
    return None

def loopclean(ldev):
    """Delete the loop mapping.
    """
    retries = 10
    while retries:
        fd = os.popen("losetup -d %s" % ldev)
        if not fd.close():
            break
        else:
            time.sleep(0.1)
            retries -= 1

def nbdsetup(pdev):
    """Sets up the nbd mapping for a qcow2 disk image.
    """
    i = 0
    while True:
        ndev = '/dev/nbd%i' % (i)
        if not os.path.exists(ndev):
            if i == 0:
                os.system("modprobe nbd > /dev/null 2>&1")
                if os.path.exists(pdev):
                    continue
            break
        i += 1
        fd = os.popen("qemu-nbd -c %s '%s' 2> /dev/null" % (ndev, pdev))
        if not fd.close():
            log.debug("qemu-nbd -c %s '%s'" % (ndev, pdev))
            return ndev
    return None

def nbdclean(ndev):
    """Delete the nbd mapping.
    """
    retries = 10
    while retries:
        fd = os.popen("qemu-nbd -d %s" % ndev)
        if not fd.close():
            break
        else:
            time.sleep(0.1)
            retries -= 1

def cleanupmapping(dev):
    """Delete the mapping.
    """
    if dev:
        if 'loop' in dev:
            loopclean(dev)
        else:
            nbdclean(dev)

def _get_partitions(dev):
    """Returns a list of dictionaries that describe the (relevant)
    partitions on the device.
    """
    found = False
    partitions = []
    # Old version of sfdisk returns Id=... New version returns type=
    pre_old = re.compile('([^ ]*)[ \\t]*:[ \\t]*start=[ \\t]*([0-9]*)[ \\t]*, size=[ \\t]*([0-9]*)[ \\t]*, Id=[ \\t]*([0-9a-fA-F]*)')
    pre = re.compile('([^ ]*)[ \\t]*:[ \\t]*start=[ \\t]*([0-9]*)[ \\t]*, size=[ \\t]*([0-9]*)[ \\t]*, type=[ \\t]*([0-9a-fA-F-]*)')
    log.debug("Examining '%s' for partitions..." % dev)
    try:
        lines = util.run(['/sbin/sfdisk', '--force', '-d', '-uS', '--', dev])
    except:
        return partitions
    for line in lines.splitlines():
        m = pre.match(line)
        if m is None:
            m = pre_old.match(line)
        if m:
            partition = {}
            pname = m.group(1)
            pstart = m.group(2)
            pid = m.group(4)
            # For now, only investigate 83 (Linux) and 6 (FAT16)
            if pid not in ('6', '83', 'ee'):
                # Check if type is GUID for a GPT partition
                try:
                    UUID(pid)
                    pid = 'ee'
                except:
                    continue
            log.debug("Found partition '%s'." % (pname,))
            partition['name'] = pname
            partition['id'] = pid
            partition['offset'] = int(pstart) * 512
            partitions.append(partition)
            found = True

    # Try '8e' (Linux LVM) scan
    if not found:
        log.debug("Try LVM scan: vdev %s" % (dev,))
        vgname = []
        fd = os.popen("pvscan | awk '$2 == \"%s\" {print $4}' 2>/dev/null" % (dev))
        for line in fd.readlines():
            line = line.strip()
            vgname = line[:]
        fd.close()
        if not vgname:
            return partitions

        log.debug("LVM Found vg: vdev %s" % (vgname))
        ret = os.system("vgchange -ay '%s' > /dev/null 2>&1" % (vgname)) >> 8
        if not ret:
            log.debug("lvscan | grep '/dev/%s'" % (vgname))
            fd = os.popen("lvscan | grep '/dev/%s'" % (vgname))
            for line in fd.readlines():
                line = line.strip()
                (t1, lvname, t2) = line.split('\'')
                pname = lvname[lvname.rfind('/')+1:]
                pname = pname.strip()
                pname = "/dev/mapper/" + vgname + "-" + pname
                partition = {}
                partition['name'] = pname
                partition['id'] = '8e'
                partition['offset'] = 0
                partition['vdev'] = vgname
                partitions.append(partition)
                log.debug("Found partition: vdev %s, pdev %s" % (vgname, pname))
            fd.close()
            log.debug("vgchange -an '%s'" % (vgname))
            os.system("vgchange -an '%s' > /dev/null 2>&1" % (vgname))
        
    return partitions

def checkAndSetDefaultBootKernel(path):
    # If a native kernel has been installed in the PV guest then the
    # menu.lst file may be defaulting to the native kernel instead of
    # the xen kernel resulting in a boot failure.
    if not path.endswith('/'):
        path = path + '/'
    if os.path.exists(path + 'vmlinuz'):
        g = None
        bootloader = path + '../etc/sysconfig/bootloader'
        try:
            loader_line = util.run(['/bin/grep', '^LOADER_TYPE', bootloader])
            if loader_line and 'grub2' in loader_line and os.path.exists(path + 'grub2/grub.cfg'):
                g = grub.GrubConf.Grub2ConfigFile(path + 'grub2/grub.cfg')
                if path.endswith('boot/'):
                    path = path.rstrip('boot/')
                menu_file = path + '/etc/default/grub'
                menu_file_field = 'GRUB_DEFAULT'
            elif os.path.exists(path + 'grub/menu.lst'):
                g = grub.GrubConf.GrubConfigFile(path + 'grub/menu.lst')
                menu_file = path + 'grub/menu.lst'
                menu_file_field = 'default'
        except:
            # Any error condition here will be ignored to allow the
            # installation to fail in the attempt to boot the native kernel.
            log.info("Warning: the grub configuration file may be defaulting to a native kernel boot")
            return
        if g:
            try:
                default = int(g._get_default())
            except:
                default = 0
            img = g.images[default]
            if img and not img.kernel or 'xen' not in img.kernel[1]:
                new_default = 0
                for img in g.images:
                    if img.kernel and 'xen' in img.kernel[1]:
                        sed_cmd = "/bin/sed --in-place 's/^%s %s/%s %s/g' %s" % (menu_file_field, default, menu_file_field, new_default, menu_file)
                        log.debug("Setting '%s' to default to the xen kernel with command: '%s'" % (menu_file, sed_cmd))
                        os.system(sed_cmd)
                        break
                    new_default = new_default + 1

class ParaVirtGuest(XenGuest):
    def __init__(self, options):
        # Set these early, in case the XenGuest.__init__ raises.
        self.kernel = ''
        self.initrd = ''
        XenGuest.__init__(self, options)
        self.runtime = {}

    def _get_runtime_images_using_pygrub(self, dev):
        _kernel = None
        _initrd = None
        # See if pygrub can decipher what is in this image.
        # Calling 'pygrub -q <image>' on a valid image copies the
        #   the kernel and initrd to /var/run/xend/boot
        # The files will be in the form, boot_kernel.<unique id> and
        #   boot_ramdisk.<unique id>
        log.debug("Using pygrub method to lookup kernel and initrd")
        fd = os.popen("pygrub -q %s 2> /dev/null" % dev)
        for line in fd.readlines():
            if 'kernel' in line and 'ramdisk' in line:
                lines = line.split('(')
                for line in lines:
                    line = line.rstrip(')')
                    if line.startswith("kernel "):
                        _kernel = line.split(" ")[1]
                    elif line.startswith("ramdisk "):
                        _initrd = line.split(" ")[1]
                if _kernel or _initrd:
                    log.debug("pygrub lookup found kernel = %s, initrd = %s" % (_kernel,_initrd))
                    # This is a post install check so remove the files copied by pygrub
                    try: os.unlink(_kernel)
                    except: pass
                    try: os.unlink(_initrd)
                    except: pass
                    break
        fd.close()
        return (_kernel, _initrd)

    def _get_paravirt_images(self, dev, partitions, copy, virtdev=None, progress=lambda f:None):
        def traildigits(strg):
            """Return the trailing digits, used to split the partition number off"""
            idx = len(strg)-1
            while strg[idx].isdigit():
                if len == 0:
                    return strg
                idx -= 1
            return strg[idx+1:]
        actions = VMDefaults[self.options.os_type].actions()
        for partition in partitions:
            # For now, only investigate 6 (FAT16) and 83 (Linux), 8e (Linux LVM)
            if partition['id'] not in ('6', '83', '8e', 'ee'):
                continue
            if partition['id'] == '8e':
                os.system("vgchange -ay '%s' > /dev/null 2>&1" % (partition['vdev']))
                dev = partition['name']
            log.debug("Loop-mounting '%s', offset %d..." % (dev, partition['offset']))
            vdisk = mount.TemporaryLoopMount(dev)
            try:
                mountpoint = vdisk.mount(options='loop,rw,offset=%d' % partition['offset'])
            except Exception as e:
                log.debug("Mount failed (%s); skipping." % str(e))
                if partition['id'] == '8e':
                    os.system("vgchange -an '%s' > /dev/null 2>&1" % (partition['vdev']))
                if 'ext4' in str(e):
                    log.debug("Mount failed because of unsupported ext4 file system.")
                    if not virtdev:
                        virtdev = dev
                    (_kernel, _initrd) = self._get_runtime_images_using_pygrub(virtdev)
                    if _kernel:
                        if partition['id'] == '8e':
                            vpart = 'LVM-' + partition['name'][partition['name'].rfind('/')+1:]
                        else:
                            vpart = traildigits(partition['name'])
                        return (vpart, _kernel, _initrd)
                if 'reiserfs' in str(e) and self.options.os_type in ['sles10', 'sled10']:
                    log.debug("Mount failed because of unsupported reiserfs file system.")
                    # Mount on SLE15 stopped working for reiserfs
                    return ("xvda2", "kernel-xen", "initrd-xen")
                continue
            try:
                (_kernel, _initrd) = actions.getRuntimeImagesFromPath(mountpoint, self.options.os_type)
                if _kernel:
                    if partition['id'] == '8e':
                        vpart = 'LVM-' + partition['name'][partition['name'].rfind('/')+1:]
                    else:
                        vpart = traildigits(partition['name'])
                    log.info("Found kernel on partition %s." % vpart)
                    # For kernel-xen. Not needed for the pvops kernel.
                    boot_kernel_check = ['sles12', 'sled12', \
                                         'sles11','sled11', \
                                         'opensuse11','opensuse12','opensuse13', \
                                         'opensuse42',]
                    if self.options.os_type in boot_kernel_check:
                        path = os.path.dirname(_kernel)
                        try:
                            checkAndSetDefaultBootKernel(mountpoint + path)
                        except:
                            pass
                    if copy:
                        # Be careful to explicitly close on failure, and close
                        # immediately after use, otherwise no guarantee these
                        # will be closed by GC before attempting unmount.
                        kfd = open(mountpoint + _kernel, 'r')
                        try:
                            ifd = None
                            if _initrd:
                                ifd = open(mountpoint + _initrd, 'r')
                        except:
                            kfd.close()
                            raise
                        try:
                            kernel = util.copy_to_temp(kfd, prefix='kernel.')
                            if not ifd:
                                initrd = None
                            else:
                                try:
                                    initrd = util.copy_to_temp(ifd, prefix='initrd.')
                                except:
                                    try: os.unlink(kernel)
                                    except: pass
                                    raise
                        finally:
                            kfd.close()
                            if ifd:
                                ifd.close()
                    else:
                        log.info("Not copying; will use kernel/initrd in-place")
                        kernel = _kernel
                        initrd = _initrd
                    log.info("kernel %s; initrd %s" % (kernel, initrd))
                    return (vpart, kernel, initrd)
            finally:
                vdisk.unmount()
                if partition['id'] == '8e':
                    log.debug("vgchange -an '%s'" % (partition['vdev']))
                    os.system("vgchange -an '%s' > /dev/null 2>&1" % (partition['vdev']))
                progress(1.0)
        progress(1.0)
        return (None, None, None)

    def _get_paravirt_install_images(self, progress=lambda f:None):
        """Retrieves the kernel and initrd from the installation source.
        Potentially a heavy-weight operation."""
        actions = VMDefaults[self.options.os_type].actions()
        source = self.options.source
        nonpae = VMDefaults[self.options.os_type].supports_32nonpae()
        if type(source) is vmdisks.Disk:
            log.debug("Installation source: %s" % source)
            if not source.exists():
                raise InstSrcError(InstSrcError.E_NOT_FOUND, details=str(source))
            vdev = source.block_attach()
            try:
                partitions = _get_partitions(vdev)
                if len(partitions) == 0:
                    mounted_disk = mount.TemporaryLoopMount(vdev)
                    try:
                        mountpoint = mounted_disk.mount()
                    except Exception as e:
                        raise InstSrcError(InstSrcError.E_BAD, details=e.args[0])
                    try:
                        (kfn, ifn) = actions.getInstallImagesFromPath(self.options, mountpoint, progress, nonpae)
                    finally:
                        mounted_disk.unmount()
                else:
                    (_, kfn, ifn) = self._get_paravirt_images(vdev, partitions, copy=True, progress=progress)
            finally:
                source.block_detach()
        else:
            (kfn, ifn) = actions.getInstallImagesFromUrl(self.options, source, progress, nonpae)
        if kfn is None:
            raise InstSrcError(InstSrcError.E_BAD, details=str(source))
        return (kfn, ifn)

    def _get_install_xml(self):
        os = "<os type='%s'" % self.options.os_type
        if self.options.install == OS_INSTALL_YES:
            if type(self.options.source) is vmdisks.Disk:
                src = 'dev:/' + self.options.source.get_vdev_str()
            else:
                src = self.options.source
            os += " source='%s'" % src
        os += '>'
        if self.initrd:
            _initrd = self.initrd
        else:
            _initrd = ""
        self._update_extra_args_with_screen_resolution()
        return [ os,
                 [ '<type>linux</type>',
                   '<kernel>%s</kernel>' % self.kernel,
                   '<initrd>%s</initrd>' % _initrd,
                   '<cmdline>%s %s</cmdline>' % (self.options._install_args, self.options.extra_args) ],
                 '</os>' ]

    def _get_runtime_xml(self):
        ramdisk = ''
        if self.runtime['initrd']:
            ramdisk = ',%s' % self.runtime['initrd']
        if self.options.verify_disk_bootable or \
           self.options.os_type == "sles15" or self.options.os_type == "sled15" or \
           self.options.os_type == "sles12" or self.options.os_type == "sled12" or \
           self.options.os_type == "opensuse13" or self.options.os_type == "opensuse42" or \
           self.options.os_type == "opensuse15" or self.options.os_type == "opensuse_tumbleweed" or \
           self.options.os_type == "rhel7":
            if os.path.exists("/usr/share/grub2/x86_64-xen/grub.xen"):
                grub_xen = "/usr/share/grub2/x86_64-xen/grub.xen"
            else:
                grub_xen = "/usr/lib/grub2/x86_64-xen/grub.xen"
            return [ "<os type='%s'>" % self.options.os_type,
                     [ '<type>linux</type>',
                       '<kernel>%s</kernel>' % grub_xen,
                       '<cmdline>%s %s</cmdline>' % (self.options._runtime_args, self.options.extra_args) ],
                     '</os>' ]
        return [ "<bootloader>%s</bootloader>" % self.options.loader,
                 "<bootloader_args>--entry=%s:%s%s</bootloader_args>" % (self.runtime['vdev'], self.runtime['kernel'], ramdisk),
                 "<os type='%s'>" % self.options.os_type,
                 [ '<type>linux</type>',
                   '<cmdline>%s %s</cmdline>' % (self.options._runtime_args, self.options.extra_args) ],
                 '</os>' ]

    def GetConfigXml(self):
        if self.isInstall:
            os_xml = self._get_install_xml()
        else:
            os_xml = self._get_runtime_xml()
            
        return [ "<domain type='xen'>",
                 self.getCommonXml(),
                 os_xml,
                 [ '<devices>',
                   [ self._get_disk_xml(),
                     self._get_network_xml(),
                     self._get_host_devices_xml(),
                     self._get_graphics_xml() ],
                   '</devices>' ],
                 '</domain>' ]
                 
    def _update_extra_args_with_screen_resolution(self):
        if self.options.screen_resolution:
            res = self.options.screen_resolution.split('x')
            if self.options.screen_resolution == '640x480':
                video_mem = '2'
            elif self.options.screen_resolution == '1280x1024':
                video_mem = '8'
            else:
                video_mem = '4'
            if caps.is_pvops_kernel():
                video_fb = "xen-fbfront.video="
            else:
                video_fb = "xenfb.video="
            if self.options.extra_args:
                index = self.options.extra_args.find("%s" % video_fb)
                if index == -1:
                    self.options.extra_args = self.options.extra_args + " %s%s,%s,%s" % (video_fb,video_mem,res[0],res[1])
            else:
                self.options.extra_args = "%s%s,%s,%s" % (video_fb,video_mem,res[0],res[1])

    def _get_install_xen(self):
        config = 'type="pv"\nkernel="%s"\n' % self.kernel
        if self.initrd:
            config += 'ramdisk="%s"\n' % self.initrd
        self._update_extra_args_with_screen_resolution()
        config += 'extra="%s %s"\n' % (self.options._install_args, self.options.extra_args)
        return config

    def _get_runtime_xen(self):
        ramdisk = ''
        if self.runtime['initrd']:
            ramdisk = ',%s' % self.runtime['initrd']
        if self.options.verify_disk_bootable or \
           self.options.os_type == "sles15" or self.options.os_type == "sled15" or \
           self.options.os_type == "sles12" or self.options.os_type == "sled12" or \
           self.options.os_type == "opensuse13" or self.options.os_type == "opensuse42" or \
           self.options.os_type == "opensuse15" or self.options.os_type == "opensuse_tumbleweed" or \
           self.options.os_type == "rhel7":
            if os.path.exists("/usr/share/grub2/x86_64-xen/grub.xen"):
                grub_xen = "/usr/share/grub2/x86_64-xen/grub.xen"
            else:
                grub_xen = "/usr/lib/grub2/x86_64-xen/grub.xen"
            return """type="pv"
#bootargs=""
#extra="%(runtime_args)s %(extra_args)s"
kernel="%(grub_arg)s"
""" % { 'runtime_args': self.options._runtime_args, 'extra_args': self.options.extra_args, 'grub_arg': grub_xen }

        if 'pygrub' in self.options.loader:
            return """type="pv"
bootloader="%(loader)s"
bootargs=""
extra="%(runtime_args)s %(extra_args)s"
""" % { 'loader': self.options.loader, 'runtime_args': self.options._runtime_args, 'extra_args': self.options.extra_args }
        else:
            return """type="pv"
bootloader="%(loader)s"
bootargs="--entry=%(vdev)s:%(kernel)s%(ramdisk)s"
extra="%(runtime_args)s %(extra_args)s"
""" % { 'loader': self.options.loader, 'vdev': self.runtime['vdev'], 'kernel': self.runtime['kernel'],
        'ramdisk': ramdisk, 'runtime_args': self.options._runtime_args, 'extra_args': self.options.extra_args }

    def GetConfigXen(self):
        if self.isInstall:
            osblob = self._get_install_xen()
        else:
            osblob = self._get_runtime_xen()

        return """%(common)s%(osblob)s%(disks)s
%(networks)s
%(host_devices)s
%(graphics)s
""" % { 'common': self.getCommonXen(), 'osblob': osblob, 'disks': self._get_disk_xen(), 'networks': self._get_network_xen(), 'host_devices': self._get_host_devices_xen(), 'graphics': self._get_graphics_xen() }

    def PreInstall(self, progress=lambda f:None):
        if self.options.install != OS_INSTALL_UPGRADE:
            XenGuest.PreInstall(self, lambda f:progress(f*0.5))
        self._cleanup_kernel()  # In case the install was retried
        (self.kernel, self.initrd) = self._get_paravirt_install_images(lambda f:progress(0.5+f*0.5))

    def Start(self):
        try:
            XenGuest.Start(self)
        finally:
            self._cleanup_kernel()
        
    def PostInstallCheck(self):
        """Probes the installed disk(s) for kernel/initrd to boot from."""
        def _get_paravirt_runtime_bootable():
            hds = [disk for disk in self.options.disks if disk.vdevtype == Disk.DEVICE_DISK]
            if hds:
                vdev = hds[0].block_attach()
                try:
                    if hds[0].has_bootsector():
                        return (hds[0].get_vdev_str(), 'grub.xen', '')
                finally:
                    hds[0].block_detach()
            raise VmError(VmError.E_NO_BOOTLOADER)

        def _get_paravirt_runtime_images():
            r = (None, None, None)
            vdev = None
            for disk in self.options.disks:
                # FIXME: skip installation source?
                try:
                    dproto = disk.get_proto()
                    if 'qcow' in str(dproto):
                        pdev = disk.get_path()
                    else:
                        vdev = disk.block_attach()
                except:
                    continue
                try:
                    try:
                        log.debug("Post install check, loopsetup...")
                        if vdev:
                            ldev = loopsetup(vdev)
                        else:
                            ldev = nbdsetup(pdev)
                        if ldev is not None:
                            partitions = _get_partitions(ldev)
                            r = self._get_paravirt_images(ldev, partitions, copy=False, virtdev=vdev)
                            cleanupmapping(ldev)
                        else:
                            partitions = _get_partitions(vdev)
                            r = self._get_paravirt_images(vdev, partitions, copy=False)
                    except OSError as ex:
                        log.debug("Probe kernel error:%d, %s" % (ex.errno, ex.strerror))
                        cleanupmapping(ldev)
                        continue
                finally:
                    if vdev:
                        disk.block_detach()
                    if r[0] or r[1]:
                        break
            if r[0] is None or r[1] is None:
                log.debug("Raise Error No Kernel r[0]:%s, r[1]:%s" % (r[0], r[1]))
                raise VmError(VmError.E_NO_KERNEL)
            if r[0][:r[0].find('-')] == 'LVM':
                return (r[0], r[1], r[2])
            else:
                return (disk.get_vdev_str() + r[0], r[1], r[2])
        if self.options.verify_disk_bootable:
            # Verify that the first hard disk is bootable
            (vdev, kernel, initrd) = _get_paravirt_runtime_bootable()
        else:
            # Probes the installed disk(s) for kernel/initrd to boot from
            (vdev, kernel, initrd) = _get_paravirt_runtime_images()
        # Success:  No turning back to retry installation now!
        util.addInstallSourceRepo(self.options.source)
        self.runtime['vdev'] = vdev
        self.runtime['kernel'] = kernel
        self.runtime['initrd'] = initrd

    def _cleanup_kernel(self):
        if self.options.preserve_on_error:
            log.info("Installation failed, not cleaning up disks because preserve_on_error is set")
            self.kernel = ''
            self.initrd = ''
            return
        if self.kernel:
            try:
                os.unlink(self.kernel)
            finally:
                self.kernel = ''
        if self.initrd:
            try:
                os.unlink(self.initrd)
            finally:
                self.initrd = ''

    def __del__(self):
        log.debug("ParaVirtGuest.__del__")
        self._cleanup_kernel()
