# Job abstraction to manage simultaneous installs
#
# Copyright 2006-2007 Novell, Inc.
# Author: Charles Coffing <ccoffing@novell.com>
#
# This software may be freely redistributed under the terms of the GNU
# general public license.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA 

import os
import signal
import sys
import tempfile
import stat

from .exceptions import *
from .log import log, init_log
from .globals import *
from . import caps
from . import msg
from . import paths
from . import util
from . import hypervisor
from . import xen_hypervisor
from . import libvirt_hypervisor
from . import console

class Job(object):
    """A Job manages the lifecycle of VM creation operation.
    
    Caller-visible attributes:
    Job.XYZ -- state constants
    job.jobid -- The unique job ID if backgrounded, else None
    job.state -- The current state in the installation process
    job.guest -- The VM instance (see xen_guest.py)
    job.guest.options -- The VM settings (see options.py)
    """

    (SETTINGS,
     PREP_INSTALL,
     INSTALL,
     POST_INSTALL,
     RUNTIME_CONF,
     RUNTIME,
     FINISHED,
     FAILED,
     CANCELED) = list(range(9))

    def __init__(self, background, upgrade=False, debug=False):
        """Creates a job instance.  Start the job via the 'run'
        method.  Most interaction with the job occur through callbacks
        supplied to the 'run' method.
        """
        self._states = (
            (Job.SETTINGS, msg.job_settings, 'settings', 0.02),
            (Job.PREP_INSTALL, msg.job_prep_install, 'prep_install', 0.80),
            (Job.INSTALL, msg.job_install, 'install', 0.02),
            (Job.POST_INSTALL, msg.job_post_install, 'post_install', 0.10),
            (Job.RUNTIME_CONF, msg.job_runtime_conf, 'runtime_conf', 0.02),
            (Job.RUNTIME, msg.job_runtime, 'runtime', 0.02),
            (Job.FINISHED, msg.job_finished, 'finish', 0.02),
        )
        self._guest = None
        self._state_num = Job.SETTINGS
        self._options = None
        self._xml_conf = self._xen_conf = None
        self._cancel = False
        self._vncport = None
        self.upgrade = upgrade
        self._upgrade_conf = None
        
        signal.signal(signal.SIGUSR1, self._handle_cancel_signal)

        self.background = background
        if self.background:
            (self.jobfd, self.jobfn) = tempfile.mkstemp(prefix=paths.jobdir)
            self.jobid = self.jobfn.split('/')[-1]
            print(self.jobid)
            sys.stdout.flush()
            def init():
                self.job_exit = None
                (_, self.logfn) = tempfile.mkstemp(prefix=paths.logdir+'/vm-install.', suffix='.log')
                self.pid = os.getpid()
                self._update_job_file()
            if not util.daemonize(cb=init):
                os._exit(0)
            hypervisor.connection.reset_connection()
            init_log(debug=debug, filename=self.logfn)
        else:
            init_log(debug=debug)
            self.pid = os.getpid()
            self.jobfd = self.jobfn = self.logfn = self.jobid = None

    def __del__(self):
        log.debug("Job.__del__")
        #signal.signal(signal.SIGUSR1, signal.SIG_DFL)
        if not self._options or not self._options.preserve_on_error or self._state_num != Job.FAILED:
            self._cleanup_conf()
        self.pid = None
        if self._state_num not in (Job.FINISHED, Job.FAILED, Job.CANCELED):
            log.info("Job exiting while in state %s; assuming canceled" % str(self._state_num))
            self.job_exit = JobError(JobError.E_CANCELED).exit_code()
            self._state_num = Job.CANCELED
        self._update_job_file()

    def _update_job_file(self):
        if not self.background:
            return
        state_num = self._state_num
        s = ''
        if self._vncport is not None:
            s += 'vnc ::%d\n' % self._vncport
        if self._options and self._options.uuid:
            s += 'uuid %s\n' % self._options.uuid
        if self._options and self._options.name:
            s += 'name %s\n' % self._options.name
        if self._options and self._options.config_dir:
            s += 'config_dir %s\n' % self._options.config_dir
        if self.job_exit is not None:
            s += 'exit %d\n' % self.job_exit
        if self.pid:
            s += 'pid %d\n' % self.pid
        if self._options and self._options.disks:
            s += 'disks '
            add_comma = False
            for disk in self._options.disks:
                if disk.get_created() is True:
                    if add_comma is True:
                        s += ','
                    s += '%s' % disk.get_path()
                    add_comma = True
            s += '\n'
        s += 'id %s\nlog %s\nstate %d\ninotify %s\n' % (self.jobid, self.logfn, state_num, self.jobfn)
        if self.jobfd is None:
            self.jobfd = os.open(self.jobfn, os.O_WRONLY | os.O_TRUNC)
        os.write(self.jobfd, s)
        os.close(self.jobfd)
        self.jobfd = None

    def _cleanup_conf(self):
        if self._state_num == Job.FINISHED:
            return
        if self._xml_conf:
            log.info("Removing configuration file %s" % self._xml_conf)
            try:
                os.unlink(self._xml_conf)
            finally:
                self._xml_conf = None
        if self._xen_conf:
            log.info("Removing configuration file %s" % self._xen_conf)
            try:
                os.unlink(self._xen_conf)
            finally:
                self._xen_conf = None

    def show_console(self):
        if (not self._options.noautoconsole) and self._ConsoleCB is not None:
            self._ConsoleCB(self)
        else:
            wait_on_me = self._show_console()
            if type(wait_on_me) is str:
                console.console.wait_for_uuid_to_exit(wait_on_me)
            elif type(wait_on_me) is int:
                (pid, status) = os.waitpid(wait_on_me, 0)

    def _show_console(self):
        """Displays a console to the user, and if appropriate, waits for the
        console to close before continuing."""
        detached = (self.state[0] != Job.INSTALL)
        if detached is True and self._options.multi_stage_install > 0:
            detached = False
        if self._options.noautoconsole:
            log.info('Not starting a console due to --no-autoconsole')
            domid = console.console.get_domid_from_uuid(self._options.uuid)
            if self._options.graphics == 'none':
                self._vncport = None
            else:
                self._vncport = console.console.get_vnc_port_from_domid(domid, timeout=10.0)
            log.debug('VNC port is %s' % self._vncport)
            if self._vncport is not None:
                self._update_job_file()
            if not detached:
                console.console.wait_for_domid_to_exit(domid, self._options.uuid)
        else:
            if detached:
                log.info('Starting a console, but not waiting for it to finish')
            else:
                log.info('Starting a console, and will wait for it to finish')
            child = None
            try_text = False
            if self._options.graphics == 'none':
                try_text = True
            else:
                if self._options.graphics_viewer == 'sdl':
                    # xm automatically created a viewer, so just wait.
                    log.debug('Viewer is sdl')
                elif self._options.graphics_viewer == 'vnc' or self._options.graphics_viewer == 'virt-viewer':
                    log.debug('Viewer is vnc')
                    try:
                        child = console.console.vnc_console(self.guest, detach=detached)
                    except NotImplementedError:
                        try_text = True
                    except RuntimeError as e:
                        self._ErrorCB(self, str(e))
            if try_text:
                log.debug('Viewer is text')
                try:
                    child = console.console.text_console(self.guest, detach=detached)
                except NotImplementedError:
                    self._ErrorCB(self, msg.display_failed)
            if child is None:
                if detached:
                    return None
                log.debug('Not detached but no child; wait for VM to exit')
                return self._options.uuid
            else:
                assert(not detached)
                log.debug('Not detached so waiting on PID')
                return child

    def run(self, GetSettingsCB, StateChangeCB=lambda job,state:None, ConsoleCB=None, \
            ErrorCB=lambda job,msg:None, ProgressCB=lambda job,f:None):
        """Starts the VM creation job.

        GetSettingsCB:  Called with (job,) to get VM's settings.  Required.
        StateChangeCB:  Called with (job, (state_N, of_M, description)) when
                        state changes.  Optional.
        ConsoleCB:  Called when console needs to be opened.  Optional.
        ErrorCB:  Called with (job, msg) when an error occurs.  Optional.
        ProgressCB:  Called with (job, fraction) as job progresses.  Fraction
                     ranges from 0.0 to 1.0.  Optional.
        """
        self._StateChangeCB = StateChangeCB
        self._GetSettingsCB = GetSettingsCB
        self._ConsoleCB = ConsoleCB
        self._ErrorCB = ErrorCB
        self._cancel = False

        def _ProgressCB(self, f):
            try:
                ProgressCB(self, f)
            except:
                # Do not trust user's progress function to not raise.  I can
                # do nothing about such an exception, and in fact lower code
                # relies on a no-throw guarantee.
                pass
        progress = util.filtered_progress(lambda f:_ProgressCB(self, f))

        try:
            try:
                done = 0.0
                for state in self._states:
                    self._state_num = state[0]
                    self._update_job_file()
                    self._StateChangeCB(self, self.state)
                    func = eval('self._state_%s' % state[2])
                    func(lambda f:progress(done+f*state[3]))
                    done += state[3]
                    if self._cancel:
                        log.debug("Job's cancel flag was set")
                        raise JobError(JobError.E_CANCELED)
                    # _state_settings sets up _options and is the first thing done.
                    if self._options and self._options.no_install == True:
                        log.info("Writing install-time configuration file - no install")
                        progress(0.5)
                        self._write_conf()
                        self.job_exit = 0
                        progress(1.0)
                        if self._options.ui == 'text':
                            sys.stdout.write(msg.job_conf_written + '\n')
                        break
            except KeyboardInterrupt:
                raise JobError(JobError.E_CANCELED)
        except Error as e:
            log.info("vminstall.Error caught in job.run: %s" % str(e))
            if self._guest:
                self._guest.FailedInstall()
                self._guest = None
            if self.background:
                self.job_exit = e.exit_code()
                self._state_num = Job.FAILED
                self._update_job_file()
            raise
        except Exception as e:
            log.info("Exception caught in job.run: %s" % str(e))
            if self._guest:
                self._guest.FailedInstall()
                self._guest = None
            if self.background:
                self.job_exit = 1
                self._state_num = Job.FAILED
                self._update_job_file()
            raise
        finally:
            if hypervisor.connection:
                del hypervisor.connection
                hypervisor.connection = None

    def _handle_cancel_signal(self, signum, frame):
        log.debug("handle_cancel_signal")
        self.do_cancel(forceful=True)

    def signal_cancel(self):
        """Attempts to cancel the job.  Not safe to call in threaded environment."""
        os.kill(self.pid, signal.SIGUSR1)

    def do_cancel(self, forceful=False):
        """Attempts to cancel the job.  Safe to call from another thread."""
        if not forceful and self._state_num == Job.FINISHED:
            return
        self._cancel = True
        # Kill VM to break out of main job loop.
        # Beware: self._guest.uuid may not exist, or may disappear.
        #         Hold reference before testing.
        guest = self._guest
        if guest:
            uuid = guest.uuid
            if uuid:
                log.debug("Destroying VM %s" % uuid)
                hypervisor.connection.destroy(uuid)

    def get_guest(self):
        return self._guest
    guest = property(get_guest)

    def get_state(self):
        return (self._state_num, len(self._states), self._states[self._state_num][1])
    state = property(get_state)

    def _state_settings(self, progress):
        log.debug("Job._state_settings")
        if not caps.is_dom0() and not caps.is_kvm() and not caps.is_qemu():
            raise XenError(XenError.E_NO_XEN)
        if os.geteuid() != 0 and not caps.is_kvm() and not caps.is_qemu():
            raise XenError(XenError.E_ROOT)

        self._options = self._GetSettingsCB(self)
        progress(0.5)

        if self._options.noautoconsole:
            if hypervisor.connection.use_libvirt is False:
                if self._options.use_xl:
                    from .xl_console import XlConsole
                    console.console = XlConsole()
                else:
                    from .xen_console import XenConsole
                    console.console = XenConsole()
            else:
                from .libvirt_console import LibVirtConsole
                console.console = LibVirtConsole()

        if self._options.full_virt:
            if caps.is_kvm() or caps.is_qemu():
                from .full_virt_kvm_guest import FullVirtKVMGuest
                self._guest = FullVirtKVMGuest(self._options)
            else:
                from .full_virt_xen_guest import FullVirtXenGuest
                self._guest = FullVirtXenGuest(self._options)
        else:
            from .ParaVirtGuest import ParaVirtGuest
            self._guest = ParaVirtGuest(self._options)

        if self._options.host_devices:
            self._guest.prepare_host_devices_for_passthrough()

        progress(1.0)

    def _write_conf(self, make_copy=False):
        def indent(nodes, depth=0):
            def process(node):
                if type(node) is str:
                    return '  '*depth+node
                return indent(node, depth+1)
            return '\n'.join(map(process, nodes))
        xml_cf = os.path.join(self.guest.options.config_dir, self.guest.options.name + '.xml')
        f = open(xml_cf, "w+")
        domXML = indent(self.guest.GetConfigXml())
        f.write(domXML)
        f.close()

        # Create a copy of the XML config in the image file location
        if make_copy is True:
            config_copy_path = paths.imagedir + self.guest.options.name
            disk = self.guest.options.disks[0]
            if disk and disk.get_vdevtype() == disk.DEVICE_DISK:
                if disk.get_proto() not in ('phy','nbd','iscsi','npiv'):
                    config_copy_path = os.path.dirname(disk.get_path())
            try:
                if not os.path.exists(config_copy_path):
                    os.makedirs(config_copy_path)
                xml_cf_copy = os.path.join(config_copy_path, self.guest.options.name + '.orig.xml')
                f = open(xml_cf_copy, "w+")
                f.write(domXML)
                f.close()
            except:
                pass

        xen_cf = os.path.join(self.guest.options.config_dir, self.guest.options.name)
        if caps.is_xen():
            f = open(xen_cf, 'w+')
            domCFG = self.guest.GetConfigXen()
            f.write(domCFG)
            f.close()

            # Create a copy of the legacy config in the image file location
            if make_copy is True:
                try:
                    xen_cf_copy = os.path.join(config_copy_path, self.guest.options.name + '.orig')
                    f = open(xen_cf_copy, 'w+')
                    f.write(domCFG)
                    f.close()
                except:
                    pass
        return (xml_cf, xen_cf)
        
    def _find_closing_brace(self, line, index, end):
        closing_brace_found = 0
        while index < end:
            if line[index] == ")" or line[index] == ">":
                closing_brace_found = 1
                index = index + 1
                break
            index += 1
        return (index, closing_brace_found)

    def _partition_line(self, line, sep):
        list = line.split(sep)
        a = b = c = ""
        idx = 0
        b = sep
        for s in list:
            if idx == 0:
                a = s 
            else:
                c = c + s
            idx = idx + 1
        return (a,b,c)

    def _parse_sxp(self, f, lines):
        kernel_found = 0
        bootloader_found = 0
        bootloader_args_found = 0
        ramdisk_found = 0
        args_found = 0
        on_reboot_found = 0

        for line in lines:
            while True:
                if not kernel_found:
                    index = line.find("(kernel ")
                    if index != -1:
                        kernel_found = 1
                        _line = line[0:index]
                        new = "           (kernel " + self.guest.kernel + ")\n"
                        if not ramdisk_found:
                            new = new + "           (ramdisk " + self.guest.initrd + ")\n"
                            ramdisk_found = 1
                        _line = _line + new
                        end = len(line)
                        (index,found) = self._find_closing_brace(line, index, end)
                        line = _line + line[index:end]
                        continue
                if not bootloader_found:
                    index = line.find("(bootloader ")
                    if index != -1:
                        bootloader_found = 1
                        end = len(line)
                        (end,found) = self._find_closing_brace(line, index, end)
                        sep = line[index:end]
                        (a,b,c) = self._partition_line(line, sep)
                        if c.strip(" \n") == "" and a.strip(" ") == "":
                            c = ""
                        if a.strip(" ") == "":
                            a = ""
                        line = a + c
                        continue
                if not bootloader_args_found:
                    index = line.find("(bootloader_args ")
                    if index != -1:
                        bootloader_args_found = 1
                        end = len(line)
                        (end,found) = self._find_closing_brace(line, index, end)
                        sep = line[index:end]
                        (a,b,c) = self._partition_line(line, sep)
                        if c.strip(" \n") == "" and a.strip(" ") == "":
                            c = ""
                        if a.strip(" ") == "":
                            a = ""
                        line = a + c
                        continue
                if not ramdisk_found:
                    index = line.find("(ramdisk ")
                    if index != -1:
                        ramdisk_found = 1
                        end = len(line)
                        (end,found) = self._find_closing_brace(line, index, end)
                        sep = line[index:end]
                        (a,b,c) = self._partition_line(line, sep)
                        if found:
                            new = "           (ramdisk " + self.guest.initrd + ")\n"
                        else:
                            new = "           (ramdisk " + self.guest.initrd + "\n"
                        line = a + new + c
                        continue
                if not args_found:
                    index = line.find("(args ")
                    if index != -1:
                        args_found = 1
                        end = len(line)
                        (end,found) = self._find_closing_brace(line, index, end)
                        sep = line[index:end]
                        (a,b,c) = self._partition_line(line, sep)
                        args_line = "           (args \' install=" + self.guest.options.source
                        if self.guest.options.extra_args:
                            args_line = args_line + " " + self.guest.options.extra_args
                        if found:
                            new = args_line + "\')\n"
                        else:
                            new = args_line + "\'\n"
                        line = a + new + c
                        continue
                if not on_reboot_found:
                    index = line.find("(on_reboot ")
                    if index != -1:
                        on_reboot_found = 1
                        end = len(line)
                        (end,found) = self._find_closing_brace(line, index, end)
                        sep = line[index:end]
                        (a,b,c) = self._partition_line(line, sep)
                        if found:
                            new = "(on_reboot destroy)" 
                        else:
                            new = "(on_reboot destroy" 
                        line = a + new + c
                        continue
                f.write(line)
                break

    def _parse_xml(self, f, lines):
        kernel_found = 0
        initrd_found = 0
        cmdline_found = 0
        on_reboot_found = 0
        os_end_found = 0
        bootloader_found = 0
        bootloader_args_found = 0
        for line in lines:
            while True:
                if not os_end_found:
                    index = line.find("</os>")
                    if index != -1:
                        os_end_found = 1
                        initrd_str = ""
                        cmdline_str = ""
                        kernel_str = ""
                        if not kernel_found:
                            kernel_str = "    <kernel>" + self.guest.kernel + "</kernel>\n"
                            kernel_found = 1
                        if not initrd_found:
                            initrd_str = "    <initrd>" + self.guest.initrd + "</initrd>\n"
                            initrd_found = 1
                        if not cmdline_found:
                            cmdline_str = "    <cmdline> install=" + self.guest.options.source
                            if self.guest.options.extra_args:
                                cmdline_str = cmdline_str + " " + self.guest.options.extra_args
                            cmdline_str = cmdline_str + "</cmdline>\n"
                            cmdline_found = 1
                        line = kernel_str + initrd_str + cmdline_str + line
                        continue
                if not kernel_found:
                    index = line.find("<kernel>")
                    if index != -1:
                        kernel_found = 1
                        _line = line[0:index]
                        line = _line + "<kernel>" + self.guest.kernel + "</kernel>\n"
                        continue
                if not initrd_found:
                    index = line.find("<initrd>")
                    if index != -1:
                        initrd_found = 1
                        end = len(line)
                        (end,found) = self._find_closing_brace(line, index, end)
                        sep = line[index:end]
                        (a,b,c) = self._partition_line(line, sep)
                        if found:
                            new = "    <initrd>" + self.guest.initrd + "</initrd>\n"
                        else:
                            new = "    <initrd>" + self.guest.initrd + "</initrd>\n"
                        line = a + new + c
                        continue
                if not cmdline_found:
                    index = line.find("<cmdline>")
                    if index != -1:
                        cmdline_found = 1
                        end = len(line)
                        (end,found) = self._find_closing_brace(line, index, end)
                        sep = line[index:end]
                        sep = line[index:end]
                        args_line = "    <cmdline> install=" + self.guest.options.source
                        if self.guest.options.extra_args:
                            args_line = args_line + " " + self.guest.options.extra_args
                        new = args_line + "</cmdline>\n"
                        line = new
                        continue
                if not on_reboot_found:
                    index = line.find("<on_reboot>")
                    if index != -1:
                        on_reboot_found = 1
                        end = len(line)
                        (end,found) = self._find_closing_brace(line, index, end)
                        sep = line[index:end]
                        (a,b,c) = self._partition_line(line, sep)
                        if found:
                            new = "<on_reboot>destroy</on_reboot>" 
                        else:
                            new = "<on_reboot>destroy</on_reboot>" 
                        line = a + new + c
                        continue
                if not bootloader_found:
                    index = line.find("<bootloader>")
                    if index != -1:
                        # We don't want this set for upgrading
                        bootloader_found = 1
                        line = ""
                        continue
                if not bootloader_args_found:
                    index = line.find("<bootloader_args>")
                    if index != -1:
                        # We don't want this set for upgrading
                        bootloader_args_found = 1
                        line = ""
                        continue
                f.write(line)
                break

    def _write_upgrade_conf(self):
        # Edit the existing guest configuration or sxp file to prepare for upgrade.
        # 1. Remove the lines that begin with (bootloader and (bootloader_args 
        # 2. Change the line that reads, (on_reboot restart) to (on_reboot destroy)
        # 3. Modify the kernel line between the parenthesis to point at the
        #    kernel copied from the network install source to dom0.
        # 4. Insert or modify the ramdisk line below the kernel line to point 
        #    to the initrd file copied from the install source to the dom0 path.
        # 5. Modify the args line between the single quotes to point at the 
        #    installation source. (args 'install=ftp://....)
        # 6. Save the upgrade configuration file.
        if not self.guest.options.upgrade_managed_vm:
            vm_name = self._options.upgrade_config
        else:
            vm_name = self.guest.options.name
        (fname, lines, bname) = util.read_config_file(vm_name, self.guest.options.upgrade_managed_vm, 1)
        if fname is None:
            log.debug('Error: Creating temporary configuration files for upgrading has failed')
            raise InstSrcError(err=InstSrcError.E_CONFIG)
        f = open(fname, "w+")
        is_xml = 0
        if lines[0].startswith("<domain type='xen'>") or lines[0].startswith("<?xml"):
            is_xml = 1

        if self._options.upgrade_managed_vm:
            if util.is_xend_running():
                self._parse_sxp(f, lines)
            else:
                self._parse_xml(f, lines)
        else:
            if is_xml:
                self._parse_xml(f, lines)
            else:
                for line in lines:
                    # This is a legacy config file format
                    if not line.strip():
                        continue
                    value = line[:-1].split('=', 1)[1]
                    value = value.strip('"')
                    vname = line[:-1].split('=', 1)[0]
                    vname = vname.strip(" ")
                    if vname == 'on_reboot':
                        line = "on_reboot=\"destroy\"\n"
                    elif vname == 'extra':
                        if self.guest.options.extra_args:
                            line = "extra=\"install=" + self.guest.options.source + " " + self.guest.options.extra_args + "\"\n"
                        else:
                            line = "extra=\"install=" + self.guest.options.source + "\"\n"
                    elif vname == 'bootloader':
                        line = "kernel=\"" + self.guest.kernel + "\"\n"
                        f.write(line)
                        line = "ramdisk=\"" + self.guest.initrd + "\"\n"
                        f.write(line)
                        continue
                    elif vname == 'bootargs':
                        continue
                    f.write(line)
        f.close()
        self.guest.options.upgrade_config = fname
        log.debug('Temporary configuration file for upgrading = %s' % fname)
        log.debug('Temporary configuration backup file for upgrading = %s' % bname)
        return fname

    def _state_prep_install(self, progress):
        if self._options.install == OS_INSTALL_NO:
            return
        self.guest.PreInstall(lambda f:progress(f*0.98))
        log.info("Writing install-time configuration file...")
        if self._options.install == OS_INSTALL_UPGRADE:
            self._upgrade_conf = self._write_upgrade_conf()
            hypervisor.connection.delete(self.guest.options.name)
        else:
            (self._xml_conf, self._xen_conf) = self._write_conf()
        progress(1.0)

    def _state_install(self, progress):
        #if self._options.install == OS_INSTALL_UPGRADE:
        #    raise JobError(JobError.E_CANCELED)
        #    return
        if self._options.install == OS_INSTALL_NO:
            return
        self.guest.Start()
        if self._options.install != OS_INSTALL_UPGRADE:
            self.guest.manage_disks()
        if not self._cancel:
            progress(0.5)
            self.show_console()
        progress(0.9)
        self.guest.Stop()
        progress(1.0)

    def _state_post_install(self, progress):
        log.info("Post install check...")
        if self._options.os_type != 'rear':
            self.guest.PostInstallCheck()
        progress(1.0)

    def _state_runtime_conf(self, progress):
        log.info("Post install configuration...")
        self.guest.PostInstallConf()
        progress(0.9)

        log.info("Writing runtime configuration file...")
        if self._options.install != OS_INSTALL_UPGRADE and self._options.os_type != 'rear':
            self._write_conf(make_copy=True)
        progress(1.0)

    def _state_runtime(self, progress):
        if self._options.norestart:
            return
        if self._options.os_type != 'rear':
            self.guest.start_from_disk()
            progress(0.5)
            self.show_console()
            while self._options.multi_stage_install > 0:
                self._options.multi_stage_install = self._options.multi_stage_install - 1
                if hypervisor.connection.wait_domain(self._options.name):
                    self.show_console()
        progress(1.0)

    def _state_finish(self, progress):
        if self._options.install == OS_INSTALL_UPGRADE:
            fn = self.guest.options.upgrade_config
            os.unlink(fn)
            os.unlink(fn + "._bak")
        self.job_exit = 0
        progress(1.0)


def _get_job_filename(job_id):
    if '/' in job_id:
        raise JobError(JobError.E_NOT_FOUND)
    return os.path.join(paths.jobdir, job_id)

def _get_status_text(job_id):
    try:
        fh = open(_get_job_filename(job_id), 'rt')
        text = fh.read()
        fh.close()
        return text
    except:
        raise JobError(JobError.E_NOT_FOUND)

def _get_status(job_id):
    text = _get_status_text(job_id)
    status = {}
    attrs = ('exit', 'id', 'inotify', 'log', 'pid', 'state', 'uuid', 'name', 'vnc', 'config_dir', 'disks')
    for attr in attrs:
        status[attr] = None
    for line in text.splitlines():
        words = line.split()
        if not words[0] in attrs:
            assert(False)
            continue
        if len(words) > 1:
            status[words[0]] = words[1]
    for attr in ('exit', 'pid', 'state'):
        if status[attr] is not None:
            status[attr] = int(status[attr])
    # Possible (although unlikely) that the state file is truncated.
    # Ensure at least 'id' is set, to avoid needless exceptions.
    status['id'] = job_id
    return status

def _try_unlink(file):
    try:
        os.unlink(file)
    except:
        pass

def _connect_tools():
    if hypervisor.connection is None:
        if not caps.is_xen():
            hypervisor.connection = libvirt_hypervisor.LibVirtHypervisor("qemu:///system")
        elif not util.is_xend_running():
            hypervisor.connection = libvirt_hypervisor.LibVirtHypervisor("xen:///")
        else:
            hypervisor.connection = xen_hypervisor.Xend(None)

def _delete(status):
    if status['name'] is not None:
        if status['config_dir'] is not None:
            _try_unlink(os.path.join(status['config_dir'], status['name']))
            _try_unlink(os.path.join(status['config_dir'], status['name'] + '.xml'))
    if status['uuid'] is not None:
        _connect_tools()
        if hypervisor.connection.domain_running(status['uuid']):
            hypervisor.connection.destroy(status['uuid'])
        hypervisor.connection.delete(status['uuid'])
    _try_unlink(_get_job_filename(status['id']))
    _try_unlink(status['log'])
    if status['disks'] is not None:
        disk_list = status['disks'].split(',')
        for disk in disk_list:
            _try_unlink(disk)

def list():
    """Returns a list of job IDs of all known backgrounded VM install jobs.
    """
    return [fn for fn in os.listdir(paths.jobdir) if len(fn) == 6]

def cancel(job_id):
    """Cancels the specified job.  If the VM is still running, it is
    stopped.  This function may return before the VM is completely
    stopped or the associated vm-install process has completely
    exited.  The job is not implicitly deleted, so that the
    cancelation progress may be observed.
    """
    status = _get_status(job_id)
    if status['pid'] is not None:
        try:
            os.kill(status['pid'], signal.SIGUSR1)
        except:
            pass
    if status['uuid'] is not None:
        _connect_tools()
        if hypervisor.connection.domain_running(status['uuid']):
            hypervisor.connection.destroy(status['uuid'])

def status(job_id):
    """Returns a dictionary that describes the specified job.
    """
    return _get_status_text(job_id)

def detach(job_id):
    """Detaches the created VM from the specified job.  The job must
    be in the FINISHED state.  The job is deleted, but the VM's
    configuration and its disks are left intact.
    """
    status = _get_status(job_id)
    if status['state'] != Job.FINISHED:
        raise JobError(JobError.E_JOB_STATE)
    _try_unlink(_get_job_filename(job_id))
    _try_unlink(status['log'])

def delete(job_id):
    """Deletes the specified job.  The job must be in one of the
    FINISHED, FAILED, or CANCELED states.  The VM's configuration is
    deleted, as are any disks that were created during the job.
    """
    status = _get_status(job_id)
    if status['state'] not in (Job.FINISHED, Job.FAILED, Job.CANCELED):
        raise JobError(JobError.E_JOB_STATE)
    _delete(status)
