# Paths
#
# Copyright 2006-2007 Novell, Inc.
# Author: Charles Coffing <ccoffing@novell.com>
#
# This software may be freely redistributed under the terms of the GNU
# general public license.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA 

import os
import libxml2
from . import globals
import vmdisks
from . import hypervisor

configdir = '/etc/xen/vm/'
gladedir = '/usr/share/vm-install/'
imagedir = None
jobdir = '/var/run/vm-install/'
logdir = '/var/log/xen/'

#-------------------------------------------------------------------------------
# Function: load_vminstall_defaults
# Parms:    xen_default
# 
# This function is used to load customization values from the /etc/default/vm-install file
#
def load_vminstall_defaults(xen_defaults=None):
    global configdir
    global imagedir
    global logdir
    default_config_file = '/etc/default/vm-install'
    if os.path.exists(default_config_file):
        f = open(default_config_file)
        lines = f.readlines()
        f.close()
        for line in lines:
            line = line.split('#')[0]
            line = line.strip()
            if not len(line) or line.startswith('#') or not line.find('=') or line == '\n':
                continue
            line=line.split('=')
            if len(line) != 2:
                continue
            item = line[0]
            value = line[1]
            if not len(item) or not len(value):
                continue
            if xen_defaults:
                if item.startswith('XEN_IMAGE_DIR'):
                    imagedir=value
                elif item.startswith('XEN_CONFIG_DIR'):
                    configdir=value
                elif item.startswith('XEN_LOG_DIR'):
                    logdir=value
                elif item.startswith('XEN_FILE_FORMAT_TYPE'):
                    drv = value + ':'
                    if drv in vmdisks.get_disk_protocols():
                        globals.default_file_type=value
            else:
                if item.startswith('KVM_IMAGE_DIR'):
                    imagedir=value
                elif item.startswith('KVM_CONFIG_DIR'):
                    configdir=value
                elif item.startswith('KVM_LOG_DIR'):
                    logdir=value
                elif item.startswith('KVM_FILE_FORMAT_TYPE'):
                    drv = value + ':'
                    if drv in vmdisks.kvm_disk_protocols:
                        globals.default_file_type=value
            if item.startswith('GRAPHICS_VIEWER'):
                if value in ('virt-viewer', 'vnc', 'sdl', 'text'):
                    globals.default_graphics_viewer=value
            elif item.startswith('NETWORK_INSTALL_SOURCE'):
                proto = value.split(':')[0] + ':'
                if value == 'pxe' or proto in ('ftp:', 'http:', 'nfs:', 'smb:'):
                    globals.default_install_source=value
            elif item.startswith('PRESERVE_ON_ERROR'):
                if value == "1":
                    globals.default_preserve_on_error=True

#-------------------------------------------------------------------------------
# Function: lookup_default_image_path
# Parms:    default_path: location found in /etc/defaults/vm-install
#           uri: connection uri to libvirt
# 
# Default image location
# 1) Use what is specified in /etc/default/vm-install
# 2) Use the 'default' storage pool from virt-manager - if it exists
# 3) Use hardcoded /var/lib/libvirt/images or /var/lib/xen/images for 'xl' only
def lookup_default_image_path(default_path, uri):
    global imagedir

    # Use the path from /etc/default/vm-install file if specified
    if default_path:
        imagedir = default_path
        if not imagedir.endswith('/'):
            imagedir = imagedir + '/'
        return

    # Need a connection to libvirt to get 'default' libvirt storage pool path
    if hypervisor.connection is None:
        from . import libvirt_hypervisor
        hypervisor.connection = libvirt_hypervisor.LibVirtHypervisor(uri)
    virConn = hypervisor.connection.getconn()

    # Use the fallback path should the lookup fail below
    imagedir='/var/lib/libvirt/images/'
    try:
        virSP = virConn.storagePoolLookupByName('default')
    except:
        return
    if virSP:
        xmlSP = virSP.XMLDesc()
        try:
            doc = libxml2.parseDoc(xmlSP)
        except:
            return
        try:
            ctxt = doc.xpathNewContext()
            poolpath = ctxt.xpathEval("/pool/target/path")
            if len(poolpath):
                imagedir = poolpath[0].getContent()
        finally:
            doc.freeDoc()
            if not imagedir.endswith('/'):
                imagedir = imagedir + '/'

#-------------------------------------------------------------------------------
# Function: load_kvm_paths
# Parms:    none
# 
# This function is used to load the kvm-specific paths into our path variables.
# In addition, we create the paths as well since kvm will not create them on
# its own.
#
def load_kvm_paths():
    global configdir
    global imagedir
    global logdir
    global jobdir

    # Load the kvm-specific paths.
    configdir = '/etc/kvm/vm/'
    imagedir = None
    logdir = '/var/log/kvm/'

    if os.geteuid() != 0:
        homedir = os.path.expanduser("~")
        configdir = homedir + '/kvm/vm/'
        logdir = homedir + '/kvm/log'
    else:
        load_vminstall_defaults()

    lookup_default_image_path(imagedir, "qemu:///system")

    # Create the kvm-specific paths.
    if not os.path.isdir(configdir):
        os.makedirs(configdir)
    if not os.path.isdir(imagedir):
        os.makedirs(imagedir)
    if not os.path.isdir(logdir):
        os.makedirs(logdir)
    if not os.path.isdir(jobdir):
        os.makedirs(jobdir)


#-------------------------------------------------------------------------------
# Function: load_xen_paths
# Parms:    none
# 
# This function is used to load the kvm-specific paths into our path variables.
# We assume that these paths are created if necessary elsewhere.
#
def load_xen_paths(options):
    global configdir
    global imagedir
    global logdir
    global jobdir

    # Load the xen-specific paths.
    configdir = '/etc/xen/vm/'
    imagedir = None
    logdir = '/var/log/xen/'

    load_vminstall_defaults(xen_defaults=1)

    if options.use_xl is True and not imagedir:
        imagedir = '/var/lib/xen/images/'
    else:
        lookup_default_image_path(imagedir, "xen:///")

    if not os.path.isdir(configdir):
        os.makedirs(configdir)
    if not os.path.isdir(imagedir):
        os.makedirs(imagedir)
    if not os.path.isdir(logdir):
        os.makedirs(logdir)
    if not os.path.isdir(jobdir):
        os.makedirs(jobdir)

#-------------------------------------------------------------------------------
# Function: load_qemu_paths
# Parms:    none
# 
# This function is used to load the qemu-specific paths into our path variables.
# In addition, we create the paths as well since qemu will not create them on
# its own.
#
def load_qemu_paths():
    global configdir
    global imagedir
    global logdir
    global jobdir

    # Load the qemu-specific paths.
    configdir = '/etc/qemu/vm/'
    imagedir = None
    logdir = '/var/log/qemu/'

    if os.geteuid() != 0:
        homedir = os.path.expanduser("~")
        configdir = homedir + '/qemu/vm/'
        logdir = homedir + '/qemu/log'
    else:
        load_vminstall_defaults()

    if imagedir is None:
        lookup_default_image_path(imagedir, "qemu:///system")

    # Create the qemu-specific paths.
    if not os.path.isdir(configdir):
        os.makedirs(configdir)
    if not os.path.isdir(imagedir):
        os.makedirs(imagedir)
    if not os.path.isdir(logdir):
        os.makedirs(logdir)
    if not os.path.isdir(jobdir):
        os.makedirs(jobdir)

